﻿//Copyright (c) 2015 Joseph W. Barker
//MIT License
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;
using ProtoBuf;

using Sandbox;
using Sandbox.Common;
using Sandbox.Common.Components;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.Engine;
using Sandbox.Game;
using Sandbox.Game.Entities;
using Sandbox.Game.Entities.Cube;
using Sandbox.Game.Entities.Character;
using Sandbox.Game.World;
using Sandbox.ModAPI;
using Ingame=Sandbox.ModAPI.Ingame;

using VRage.Game;
using VRage.ObjectBuilders;
using VRage.Utils;
using IngameV=VRage.Game.ModAPI.Ingame;
using VRageMath;

namespace InterGridComms
{
    //prior=999 so that GetObjectBuilder happens before MyScriptManager.GetObjectBuilder
    [MySessionComponentDescriptor(MyUpdateOrder.BeforeSimulation,999)]
    public class IGC : MySessionComponentBase
    {
        public const string SAVE_KEY="InterGridComms";
        public const string CTRL_PREFIX="IGC";
        public const string CTRL_SENDMSG=CTRL_PREFIX+".send";
        public const string CTRL_RECVMSG=CTRL_PREFIX+".recv";
        public const string CTRL_OPENCHANNELS=CTRL_PREFIX+".listen";
        public const string CTRL_PBCALLBACK=CTRL_PREFIX+".callback";
        public const string CTRL_PBARG=CTRL_PREFIX+".cbarg";
        public const string CTRL_UNIQUE=CTRL_PREFIX+".unique";
        public const string CTRL_SCAN=CTRL_PREFIX+".scan";
        public const string CTRL_BCASTMSG=CTRL_PREFIX+".broadcast";

        //Limit the size of an antenna's message queue
        public const int MSG_QUEUE_LIMIT=64*1024;

        public const int CHAN_NAME_LIMIT=128;

        //Have to hardcode this because we don't have access to the ingame value
        public const float PLAYER_RADIO_RADIUS=200f;

        //When searching for objects, check a little further out to account for
        //movement between updates
        public const float RADIO_FUDGE_FACTOR=104.5f;

        public static readonly int DEBUG;

        public static IGC Instance { get; private set; }
        
        public static readonly MyLog Log = new MyLog();

        public long clock { get; private set; }

        private long save_state_clear=-1;

        //Collection of antennas that are listening for messages
        public class antInfo_t
        {
            //The antenna
            public IMyTerminalBlock ant;
            public string unique;

            //Messages waiting to be read
            public Queue<string> msgs = new Queue<string>();
            public int msgsSize=0;

            //Channels that this antenna is listening on
            public string[] channels=new string[0];
            public string channels_src;

            //PB to trigger when new messages come in
            public Ingame.IMyProgrammableBlock pbcb=null;
            public string pbarg=String.Empty;

            public antInfo_t(IMyTerminalBlock ant)
            {
                this.ant=ant;

                unique=IGC.encode_entityid(ant);
            }
        };
        private Dictionary<IMyTerminalBlock,antInfo_t> antInfo=new Dictionary<IMyTerminalBlock,antInfo_t>();

        //Collection of channels that atleast one antenna is listening to
        public class channel_t
        {
            public HashSet<antInfo_t> listeners=new HashSet<antInfo_t>();
        };
        private Dictionary<string,channel_t> chanInfo=new Dictionary<string,channel_t>();

        //Custom message that certain blocks can pass during scan
        private Dictionary<IMyEntity,string> broadcastMessages=new Dictionary<IMyEntity,string>();

        //Set of PBs that have already been triggered this update
        private HashSet<antInfo_t> needsPBTriggered=new HashSet<antInfo_t>();
        private HashSet<antInfo_t> npbtOld=new HashSet<antInfo_t>();

        public abstract class radioEnt_t
        {
            public IMyEntity ent { get; private set; }
            public IMyCubeGrid cg;

            public HashSet<radioEnt_t> connectedTo=new HashSet<radioEnt_t>();
            public HashSet<radioEnt_t> connectedFrom=new HashSet<radioEnt_t>();
            public bool wasOperatingAtLastUpdate = false;
            public long next_update_time=0;

            public HashSet<IMyTerminalBlock> connMap=new HashSet<IMyTerminalBlock>();
            public long last_map_update=0;

            public HashSet<radioEnt_t> scanMap=new HashSet<radioEnt_t>();
            public long last_scan_update=0;

            public abstract void update(long clock);
            //public abstract void updateTo(radioEnt_t other);

            public radioEnt_t(IMyEntity ent)
            {
                this.ent=ent;
                if(ent is IMyCubeBlock)
                    cg=((IMyCubeBlock)ent).CubeGrid;
            }
        }
        private Dictionary<IMyEntity,radioEnt_t> radioEnts=new Dictionary<IMyEntity,radioEnt_t>();
        private Dictionary<IngameV.IMyCubeGrid,HashSet<radioEnt_t>> grid2blks=new Dictionary<IngameV.IMyCubeGrid,HashSet<radioEnt_t>>();

        private PrioritySequence<radioEnt_t> updateSeq=new PrioritySequence<radioEnt_t>();

        //////////////////////////////////////////////////////////////////////////////

        public override void BeforeStart()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - BeforeStart");
            object[] state;
            if(((IMyUtilities)MyAPIUtilities.Static).GetVariable<object[]>(SAVE_KEY,out state) && state!=null)
                LoadState(state);
            save_state_clear=1;            
        }

        public override MyObjectBuilder_SessionComponent GetObjectBuilder()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetObjectBuilder");
            //Note, we have carefully arranged (via priorities,etc) that this runs
            //before MyScriptManager.GetObjectBuilder (which owns these variables)
            MyAPIGateway.Utilities.SetVariable<object[]>(SAVE_KEY,
                SaveState() );
            save_state_clear=30;
            return base.GetObjectBuilder();
        }

        private object[] SaveState()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SaveState");
            //Save antenna info
            object[] stateAI=new object[antInfo.Count];
            int i=0;
            foreach(var kv in antInfo)
            {
                var ai=kv.Value;

                var msgs=new string[ai.msgs.Count];
                ai.msgs.CopyTo(msgs,0);

                var msgs2=new object[msgs.Length];
                for(int j=0;j<msgs.Length;++j)
                    msgs2[j]=msgs[j].Split('\0');

                stateAI[i++]=new object[] {
                    ai.ant.EntityId,
                    ai.channels,
                    msgs2,
                    (ai.pbcb!=null)?ai.pbcb.EntityId:(long)0,
                    ai.pbarg
                };
            }

            //Save broadcast messages
            object[] stateBCM=new object[broadcastMessages.Count];
            i=0;
            foreach(var kv in broadcastMessages)
            {
                stateBCM[i++]=new object[] {
                    kv.Key.EntityId,
                    kv.Value
                };
            }

            object[] state=new object[] { 1, stateAI, stateBCM };
            return state;
        }

        private void LoadState(object[] state)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - LoadState");
            var meh=(IMyEntities)new MyEntitiesHelper_ModAPI(); //Have to do this because stupid game hasn't created it yet
            var sb=new StringBuilder();

            if((int)state[0]!=1)
                throw new System.InvalidOperationException("InterGridComms: Unrecognized save state version");

            //Load antenna info
            object[] stateAI=(object[])state[1];
            for(int i=0;i<stateAI.Length;++i)
            {
                object[] si=(object[])stateAI[i];

                long id=(long)si[0];
                IMyEntity ent;
                if(!meh.TryGetEntityById(id,out ent) ||
                   !(ent is Ingame.IMyRadioAntenna)    )
                    continue;
                IMyTerminalBlock ant=(IMyTerminalBlock)ent;

                sb.Clear();
                foreach(string s in (string[])si[1])
                {
                    sb.Append(s);
                    sb.Append('\0');
                }
                sb.Remove(sb.Length-1,1);
                string channels_src=sb.ToString();
                SetChannels(ant,channels_src);

                antInfo_t ai=antInfo[ant];
                ai.msgsSize=0;
                foreach(var o in (object[])si[2])
                {
                    sb.Clear();
                    foreach(string s in (string[])o)
                    {
                        sb.Append(s);
                        sb.Append('\0');
                    }
                    sb.Remove(sb.Length-1,1);

                    string msg=sb.ToString();
                    ai.msgs.Enqueue(msg);
                    ai.msgsSize+=msg.Length;
                }

                long pbid=(long)si[3];
                ai.pbarg=(string)si[4];
                if(pbid!=0 && meh.TryGetEntityById(pbid,out ent) &&
                   (ent is Ingame.IMyProgrammableBlock) &&
                   ((IMyCubeBlock)ent).CubeGrid==ai.ant.CubeGrid   )
                {
                    ai.pbcb=(Ingame.IMyProgrammableBlock)ent;
                    if(ai.msgs.Count>0)
                        needsPBTriggered.Add(ai);
                }
            }

            //Load broadcast messages
            object[] stateBCM=(object[])state[2];
            for(int i=0;i<stateBCM.Length;++i)
            {
                object[] si=(object[])stateBCM[i];

                long id=(long)si[0];
                IMyEntity ent;
                if(!meh.TryGetEntityById(id,out ent) ||
                   !( ent is Ingame.IMyRadioAntenna || ent is Ingame.IMyBeacon) )
                    continue;

                broadcastMessages.Add(ent,(string)si[1]);
            }
        }

        //////////////////////////////////////////////////////////////////////////////

        public override void UpdateBeforeSimulation()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UpdateBeforeSimulation");
            ++clock; //New tick, update clock
            
            //////////////////////////////////
            //Initialization that needs to happen on first frame
            //(usually because MyAPIGateway isn't setup before now)
            if(clock==1)
            {
                setupBlockProperties();
            }
            
            //////////////////////////////////
            //Clear save state to save memory
            if((--save_state_clear)==0)
                MyAPIGateway.Utilities.SetVariable<object[]>(SAVE_KEY,null);

            //////////////////////////////////
            //Update known antennas/objects
if(DEBUG>1) Log.WriteLine("InterGridComms: radioEnts.Count="+radioEnts.Count.ToString());
            while(updateSeq.LowestPrio <= clock)
            {
                radioEnt_t re=updateSeq.Next();

                if(isEntityVirtual(re.ent))
                {
if(DEBUG>1) Log.WriteLine("InterGridComms: Found virtual block \""+((IMyTerminalBlock)re.ent).CustomName+"\", "+re.ent.EntityId.ToString());
                    if(re.ent is IMyTerminalBlock)
                        UnregisterBlock((IMyTerminalBlock)re.ent);
                    continue;
                }
if(DEBUG>1)
if(re.cg!=null && re.cg!=((IMyCubeBlock)re.ent).CubeGrid)
Log.WriteLine("InterGridComms: Cube grid changed! \""+((IMyTerminalBlock)re.ent).CustomName+"\", "+re.ent.EntityId.ToString());

                long old=re.next_update_time;
                re.update(clock);

                updateSeq.Add(re.next_update_time,re);
            }

            //////////////////////////////////
            //Run PBs that have been setup as callbacks

            //First, swap in a new empty collection so that callbacks
            //triggered while we are running PBs are registered
            var t=npbtOld;
            npbtOld=needsPBTriggered;
            needsPBTriggered=t;
            needsPBTriggered.Clear();

            foreach(var ai in npbtOld)
            {
                if(ai.msgs.Count==0 || ai.pbcb==null) continue;

                if(!ai.pbcb.TryRun(ai.pbarg))
                    needsPBTriggered.Add(ai); //Try again
            }
            npbtOld.Clear();
        }

        //////////////////////////////////////////////////////////////////////////////

        public void RegisterBlock(IMyTerminalBlock blk)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - RegisterBlock");
            if(radioEnts.ContainsKey(blk))
            {
                Log.WriteLine("InterGridComms: Attempted to reregister block \""+blk.CustomName+"\"");
                return;
            }

            radioEnt_t re;
            if(blk is Ingame.IMyRadioAntenna)
                re=new radioEnt_radio_t(blk);
            else if(blk is IMyLaserAntenna)
                re=new radioEnt_laser_t(blk);
            else if(blk is Ingame.IMyBeacon)
                re=new radioEnt_beacon_t(blk);
            else
                return; //?
            radioEnts.Add(blk,re);

            re.next_update_time=clock-MyUtils.GetRandomInt(0, 55);
            updateSeq.Add(re.next_update_time,re);

            HashSet<radioEnt_t> g2b;
            if(!grid2blks.TryGetValue(blk.CubeGrid,out g2b))
            {
                g2b=new HashSet<radioEnt_t>();
                grid2blks.Add(blk.CubeGrid,g2b);
            }
            g2b.Add(re);

            re.ent.OnClose += IMyTerminalBlock_OnClose;
if(DEBUG>1) Log.WriteLine("InterGridComms: Registered block \""+blk.CustomName+"\" ("+blk.EntityId.ToString()+"), cubegrid "+((IMyEntity)blk.CubeGrid).EntityId.ToString());
        }

        public void UnregisterBlock(IMyTerminalBlock blk)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UnregisterBlock");
            radioEnt_t re;
            if(!radioEnts.TryGetValue(blk, out re)) return;

            updateSeq.Remove(re.next_update_time,re);

            foreach(radioEnt_t other in re.connectedTo)
                other.connectedFrom.Remove(re);
            re.connectedTo=null; //Cause crash if re somehow sticks around
            foreach(radioEnt_t other in re.connectedFrom)
                other.connectedTo.Remove(re);
            re.connectedFrom=null;

            radioEnts.Remove(blk);

            HashSet<radioEnt_t> g2b;
            if(grid2blks.TryGetValue(blk.CubeGrid, out g2b))
            {
                g2b.Remove(re);
                if(g2b.Count==0)
                    grid2blks.Remove(blk.CubeGrid);
            }

if(DEBUG>1) Log.WriteLine("InterGridComms: Unregistered block \""+blk.CustomName+"\" ("+blk.EntityId.ToString()+")");
        }

        public void UnregisterGrid(IMyCubeGrid cg)
        {
            grid2blks.Remove(cg);
        }

        public void RegisterCharacter(IMyEntity ply)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - RegisterCharacter");
            if(radioEnts.ContainsKey(ply))
            {
                Log.WriteLine("InterGridComms: Attempted to reregister block \""+ply.DisplayName+"\"");
                return;
            }

            radioEnt_t re=new radioEnt_character_t(ply);
            radioEnts.Add(ply,re);

            re.next_update_time=clock-MyUtils.GetRandomInt(0, 55);
            updateSeq.Add(re.next_update_time,re);

            re.ent.OnClose += MyCharacter_OnClose;
if(DEBUG>1) Log.WriteLine("InterGridComms: Registered character \""+ply.DisplayName+"\" ("+ply.EntityId.ToString()+")");
        }

        public void UnregisterCharacter(IMyEntity ply)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UnregisterCharacter");
            radioEnt_t re;
            if(!radioEnts.TryGetValue(ply, out re)) return;

            updateSeq.Remove(re.next_update_time,re);

            foreach(radioEnt_t other in re.connectedTo)
                other.connectedFrom.Remove(re);
            re.connectedTo=null; //Cause crash if re somehow sticks around
            foreach(radioEnt_t other in re.connectedFrom)
                other.connectedTo.Remove(re);
            re.connectedFrom=null;

            radioEnts.Remove(ply);

if(DEBUG>1) Log.WriteLine("InterGridComms: Unregistered character \""+ply.DisplayName+"\" ("+ply.EntityId.ToString()+")");
        }

        public static void IMyTerminalBlock_OnClose(IMyEntity ent)
        { if(Instance!=null) Instance.UnregisterBlock((IMyTerminalBlock)ent); }

        public static void MyCharacter_OnClose(IMyEntity ent)
        { if(Instance!=null) Instance.UnregisterCharacter(ent); }

        internal static bool radioEnts_TryGetValue(IMyTerminalBlock x, out radioEnt_t re)
        { return Instance.radioEnts.TryGetValue(x, out re); }

        internal static bool grid2blks_TryGetValue(IMyCubeGrid x, out HashSet<radioEnt_t> lst)
        { return Instance.grid2blks.TryGetValue(x, out lst); }

        public static bool isEntityVirtual(IMyEntity ent)
        { return ent.Transparent; /* Ugh, is this really the only way to tell if a block is a projection or paste-in-progress? */ }

        public class radioEnt_radio_t : radioEnt_t
        {
            public radioEnt_radio_t(IMyEntity ent) : base(ent) { }

            public override void update(long clock)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_radio_t::update");
if(DEBUG>2) Log.WriteLine("InterGridComms: Updating antenna "+ent.EntityId.ToString()+", \""+((IMyTerminalBlock)ent).CustomName+"\", "+ent.GetPosition().ToString());
                //Start over from scratch
                foreach(radioEnt_t other in connectedTo)
                    other.connectedFrom.Remove(this);
                connectedTo.Clear();

                //Find all other antennas in range
                var ant=(Ingame.IMyRadioAntenna)ent;
                var bsd=new BoundingSphereD(ant.GetPosition(),ant.Radius+RADIO_FUDGE_FACTOR);
                foreach(var ent2 in MyAPIGateway.Entities.GetEntitiesInSphere(ref bsd))
                {
                    var cg=ent2 as MyCubeGrid;
                    if(cg==null) continue;
                    if((IngameV.IMyCubeGrid)cg==ant.CubeGrid) continue;

if(DEBUG>2) Log.WriteLine("InterGridComms:    Found cubegrid "+cg.EntityId.ToString()+", "+((IngameV.IMyCubeGrid)cg).GetPosition().ToString());
                    updateFromCubeGrid(cg);
                }

                //Track operating so people don't cheat by only powering on antenna
                //for very short periods (i.e. only tick when sending).
                wasOperatingAtLastUpdate = ant.IsWorking && ant.IsFunctional;

                //Because the above is expensive, only update this once a second
                //(with a random variation, to help spread updates over time)
                next_update_time=clock+MyUtils.GetRandomInt(55, 65+1);
            }

            private void updateFromCubeGrid(IMyCubeGrid cg)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_radio_t::update");
                HashSet<radioEnt_t> lst;
                if(!grid2blks_TryGetValue(cg, out lst))
                    return; //No interesting blocks on this grid

                foreach(var re in lst)
                    if(re.ent is Ingame.IMyRadioAntenna)
                        updateFromAntenna(re);
            }

            private void updateFromAntenna(radioEnt_t other)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_radio_t::update");
if(DEBUG>2) Log.WriteLine("InterGridComms:       Checking antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                var ant=(Ingame.IMyRadioAntenna)ent;
                if(Vector3D.Distance(ent.GetPosition(),other.ent.GetPosition())>ant.Radius+RADIO_FUDGE_FACTOR)
                {
if(DEBUG>2) Log.WriteLine("InterGridComms:          Too far");
                    return;
                }

if(DEBUG>2) Log.WriteLine("InterGridComms:       Added antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                connectedTo.Add(other);
                other.connectedFrom.Add(this);
            }
        }

        public class radioEnt_laser_t : radioEnt_t
        {
            public radioEnt_laser_t(IMyEntity ent) : base(ent) { }

            public override void update(long clock)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_laser_t::update");
if(DEBUG>1) Log.WriteLine("InterGridComms: Updating laser "+ent.EntityId.ToString()+", \""+((IMyTerminalBlock)ent).CustomName+"\"");
                //Start over from scratch
                {
                    foreach(radioEnt_t other in connectedTo)
                        other.connectedFrom.Remove(this);
                    connectedTo.Clear();
                }

                //Find laser on the other end
                var las=(IMyLaserAntenna)ent;
                var oLas=las.GetOther();
                if(oLas!=null)
                {
                    radioEnt_t other;
                    if(radioEnts_TryGetValue((IMyTerminalBlock)oLas, out other))
                    {
                        connectedTo.Add(other);
                        other.connectedFrom.Add(this);
if(DEBUG>1) Log.WriteLine("InterGridComms:   Connected to "+other.ent.EntityId.ToString()+", \""+((IMyTerminalBlock)other.ent).CustomName+"\"");
                    }
                }
else if(DEBUG>1) Log.WriteLine("InterGridComms:   Not connected to anything");

                wasOperatingAtLastUpdate = (oLas!=null);

                next_update_time=clock+MyUtils.GetRandomInt(55, 65+1);
            }
        }

        public class radioEnt_beacon_t : radioEnt_t
        {
            public radioEnt_beacon_t(IMyEntity ent) : base(ent) { }

            public override void update(long clock)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_laser_t::update");
if(DEBUG>2) Log.WriteLine("InterGridComms: Updating beacon "+ent.EntityId.ToString()+", \""+((IMyTerminalBlock)ent).CustomName+"\", "+ent.GetPosition().ToString());
                //Start over from scratch
                if(connectedFrom.Count>0)
                    throw new System.Exception("Beacon has connectedFrom");
                foreach(radioEnt_t other in connectedTo)
                    other.connectedFrom.Remove(this);
                connectedTo.Clear();

                //Find all other antennas in range
                var bcn=(Ingame.IMyBeacon)ent;
                var bsd=new BoundingSphereD(bcn.GetPosition(),bcn.Radius+RADIO_FUDGE_FACTOR);
                foreach(var ent2 in MyAPIGateway.Entities.GetEntitiesInSphere(ref bsd))
                {
                    var cg=ent2 as MyCubeGrid;
                    if(cg==null) continue;
                    if((IngameV.IMyCubeGrid)cg==bcn.CubeGrid) continue;

if(DEBUG>2) Log.WriteLine("InterGridComms:    Found cubegrid "+cg.EntityId.ToString()+", "+((IngameV.IMyCubeGrid)cg).GetPosition().ToString());
                    updateFromCubeGrid(cg);
                }

                //Track operating so people don't cheat by only powering on antenna
                //for very short periods (i.e. only tick when sending).
                wasOperatingAtLastUpdate = bcn.IsWorking && bcn.IsFunctional;

                //Because the above is expensive, only update this once a second
                //(with a random variation, to help spread updates over time)
                next_update_time=clock+MyUtils.GetRandomInt(55, 65+1);
            }

            private void updateFromCubeGrid(IMyCubeGrid cg)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_laser_t::updateFromCubeGrid");
                HashSet<radioEnt_t> lst;
                if(!grid2blks_TryGetValue(cg, out lst))
                    return; //No interesting blocks on this grid

                foreach(var re in lst)
                    if(re.ent is Ingame.IMyRadioAntenna)
                        updateFromAntenna(re);
            }

            private void updateFromAntenna(radioEnt_t other)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_laser_t::updateFromAntenna");
if(DEBUG>2) Log.WriteLine("InterGridComms:       Checking antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                var bcn=(Ingame.IMyBeacon)ent;
                if(Vector3D.Distance(ent.GetPosition(),other.ent.GetPosition())>bcn.Radius+RADIO_FUDGE_FACTOR)
                {
if(DEBUG>2) Log.WriteLine("InterGridComms:          Too far");
                    return;
                }

if(DEBUG>2) Log.WriteLine("InterGridComms:       Added antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                connectedTo.Add(other);
                other.connectedFrom.Add(this);
            }
        }

        public class radioEnt_character_t : radioEnt_t
        {
            public radioEnt_character_t(IMyEntity ent) : base(ent) { }

            public override void update(long clock)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_character_t::update");
if(DEBUG>2) Log.WriteLine("InterGridComms: Updating character "+ent.EntityId.ToString()+", \""+ent.DisplayName+"\", "+ent.GetPosition().ToString());
                //Start over from scratch
                if(connectedFrom.Count>0)
                    throw new System.Exception("Character has connectedFrom");
                foreach(radioEnt_t other in connectedTo)
                    other.connectedFrom.Remove(this);
                connectedTo.Clear();
                
                //Find all other antennas in range
                var bsd=new BoundingSphereD(ent.GetPosition(),PLAYER_RADIO_RADIUS+RADIO_FUDGE_FACTOR);
                foreach(var ent2 in MyAPIGateway.Entities.GetEntitiesInSphere(ref bsd))
                {
                    var cg=ent2 as MyCubeGrid;
                    if(cg==null) continue;
if(DEBUG>2) Log.WriteLine("InterGridComms:    Found cubegrid "+cg.EntityId.ToString()+", "+((IngameV.IMyCubeGrid)cg).GetPosition().ToString());
                    updateFromCubeGrid(cg);
                }

/* MyRadioBroadcaster not accessible :(
                    MyDataBroadcaster mdb;
                    ((IMyComponentOwner<MyDataBroadcaster>)other.ent).GetComponent(out mdb);
                    if(mdb==null || !((MyRadioBroadcaster)mdb.Enabled) continue;
*/
                var ob=(MyObjectBuilder_Character)ent.GetObjectBuilder(); //Expensive :(
                wasOperatingAtLastUpdate = ob.EnableBroadcasting;

                //Because the above is expensive, only update this once a second
                //(with a random variation, to help spread updates over time)
                next_update_time=clock+MyUtils.GetRandomInt(55, 65+1);
            }

            private void updateFromCubeGrid(IMyCubeGrid cg)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_character_t::updateFromCubeGrid");
                HashSet<radioEnt_t> lst;
                if(!grid2blks_TryGetValue(cg, out lst))
                    return; //No interesting blocks on this grid

                foreach(var re in lst)
                    if(re.ent is Ingame.IMyRadioAntenna)
                        updateFromAntenna(re);
            }

            private void updateFromAntenna(radioEnt_t other)
            {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - radioEnt_character_t::updateFromAntenna");
if(DEBUG>2) Log.WriteLine("InterGridComms:       Checking antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                if(Vector3D.Distance(ent.GetPosition(),other.ent.GetPosition())>PLAYER_RADIO_RADIUS+RADIO_FUDGE_FACTOR)
                {
if(DEBUG>2) Log.WriteLine("InterGridComms:          Too far");
                    return;
                }

if(DEBUG>2) Log.WriteLine("InterGridComms:       Added antenna "+other.ent.EntityId.ToString()+", "+other.ent.GetPosition().ToString());
                connectedTo.Add(other);
                other.connectedFrom.Add(this);
            }
        }

        private Queue<radioEnt_t> ucm_openset=new Queue<radioEnt_t>();
        private HashSet<IngameV.IMyCubeGrid> ucm_cg=new HashSet<IngameV.IMyCubeGrid>();
        private void UpdateConnMap(radioEnt_t re)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UpdateConnMap");
            if(re.last_map_update + 60 > clock)
                return;
if(DEBUG>1) Log.WriteLine("InterGridComms: Updating connectivity map for \""+((IMyTerminalBlock)re.ent).CustomName+"\", "+re.ent.EntityId.ToString());

            re.connMap.Clear();
            ucm_openset.Clear();
            ucm_cg.Clear();

            long owner=((IMyTerminalBlock)re.ent).OwnerId;
            re.connMap.Add((IMyTerminalBlock)re.ent);
            ucm_openset.Enqueue(re);
            while(ucm_openset.Count>0)
            {
                radioEnt_t other=ucm_openset.Dequeue();
                var oAnt=other.ent as Ingame.IMyRadioAntenna;
                var oLas=other.ent as IMyLaserAntenna;

                if(!other.wasOperatingAtLastUpdate) continue;

                //Add blocks connected by radio/laser
                if(oAnt!=null)
                {
if(DEBUG>1) Log.WriteLine("InterGridComms:   Expanding antenna \""+oAnt.CustomName+"\", "+other.ent.EntityId.ToString());
                    if(!oAnt.IsBroadcasting) continue; //Can't relay if not bcasting

                    foreach(var con in other.connectedTo)
                    {
                        var cAnt=con.ent as Ingame.IMyRadioAntenna;
                        if(cAnt==null) continue;
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+cAnt.CustomName+"\", "+con.ent.EntityId.ToString());

                        if(re.connMap.Contains((IMyTerminalBlock)con.ent)) continue;

                        if(Vector3D.Distance(oAnt.GetPosition(),cAnt.GetPosition()) > oAnt.Radius) continue;

                        if(cAnt.GetUserRelationToOwner(owner) == MyRelationsBetweenPlayerAndBlock.Enemies)
                            continue;

                        //Looks ok
                        re.connMap.Add((IMyTerminalBlock)con.ent);
                        ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                    }
                }
                else if(oLas!=null)
                {
if(DEBUG>1) Log.WriteLine("InterGridComms:   Expanding laser \""+((IMyTerminalBlock)oLas).CustomName+"\", "+other.ent.EntityId.ToString());
                    if(other.connectedTo.Count>0)
                    {
                        radioEnt_t con;
                        {
                            var en=other.connectedTo.GetEnumerator();
                            en.MoveNext();
                            con=en.Current;
                        }

                        var cLas=con.ent as IMyLaserAntenna;
                        if(cLas==oLas.GetOther())
                        {
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+((IMyTerminalBlock)cLas).CustomName+"\", "+con.ent.EntityId.ToString());

                            if(!re.connMap.Contains(cLas))
                            {
                                re.connMap.Add(cLas);
                                ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                            }
                        }
                    }
                }
                else
                    //Other types can't pass signals
                    continue;

                //Add blocks on same grid
                var other_cg=((IMyCubeBlock)other.ent).CubeGrid;
                if(ucm_cg.Contains(other_cg)) //No need to check same grid twice
                {
if(DEBUG>1) Log.WriteLine("InterGridComms:     SameGrid - already checked");
                    continue;
                }
                
                //TODO: var gts=MyAPIGateway.TerminalActionsHelper.GetTerminalSystemForGrid(other_cg);
                foreach(var con in grid2blks[other_cg])
                {
                    if(!(con.ent is Ingame.IMyRadioAntenna)&&!(con.ent is IMyLaserAntenna))
                        continue;
if(DEBUG>1) Log.WriteLine("InterGridComms:     SameGrid \""+((IMyTerminalBlock)con.ent).CustomName+"\", "+con.ent.EntityId.ToString());

                    if(re.connMap.Contains((IMyTerminalBlock)con.ent)) continue;

                    if(((IMyCubeBlock)con.ent).GetUserRelationToOwner(owner) == MyRelationsBetweenPlayerAndBlock.Enemies)
                        continue;

                    re.connMap.Add((IMyTerminalBlock)con.ent);
                    ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                }
                ucm_cg.Add(other_cg);
            }
            re.last_map_update = clock;
        }

        //Note this is very similar to UpdateConnMap, but now we are checking
        //reception capability
        private void UpdateScanMap(radioEnt_t re)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UpdateScanMap");
            if(re.last_scan_update + 60 > clock)
                return;
if(DEBUG>1) Log.WriteLine("InterGridComms: Updating scan map for \""+((IMyTerminalBlock)re.ent).CustomName+"\", "+re.ent.EntityId.ToString());

            re.scanMap.Clear();
            ucm_openset.Clear();
            ucm_cg.Clear();

            long owner=((IMyTerminalBlock)re.ent).OwnerId;
            re.scanMap.Add(re);
            ucm_openset.Enqueue(re);
            while(ucm_openset.Count>0)
            {
                radioEnt_t other=ucm_openset.Dequeue();
                if(!other.wasOperatingAtLastUpdate) continue;

                if(other.ent is IMyTerminalBlock &&
                   ((IMyTerminalBlock)other.ent).GetUserRelationToOwner(owner) == MyRelationsBetweenPlayerAndBlock.Enemies)
                   continue;

                //Add blocks connected by radio/laser
                var oAnt=other.ent as Ingame.IMyRadioAntenna;
                var oLas=other.ent as IMyLaserAntenna;
                if(oAnt!=null)
                {
if(DEBUG>1) Log.WriteLine("InterGridComms:   Expanding antenna \""+oAnt.CustomName+"\", "+other.ent.EntityId.ToString());
                    bool noRelay=!oAnt.IsBroadcasting; //Can't relay if not bcasting

                    foreach(var con in other.connectedFrom)
                    {
                        if(re.scanMap.Contains(con)) continue;

                        var cAnt=con.ent as Ingame.IMyRadioAntenna;
                        var cBcn=con.ent as Ingame.IMyBeacon;

                        float radius;
                        if(cAnt!=null)
                        {
                            radius=cAnt.Radius;
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+cAnt.CustomName+"\", "+con.ent.EntityId.ToString());
                        }
                        else if(cBcn!=null)
                        {
                            radius=cBcn.Radius;
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+cBcn.CustomName+"\", "+con.ent.EntityId.ToString());
                        }
                        else if(con.ent is IMyCharacter)
                        {
                            var cPly=MyAPIGateway.Players.GetPlayerControllingEntity(con.ent);
                            radius=PLAYER_RADIO_RADIUS; //No choice but to hard code, since MyRadioBroadcaster is blocked
                            if(cPly==null) continue; //No player controlling
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+con.ent.DisplayName+"\", "+con.ent.EntityId.ToString());
                        }
                        else
                            continue;

                        if(Vector3D.Distance(oAnt.GetPosition(),con.ent.GetPosition()) > radius) continue;

                        //Looks ok
                        re.scanMap.Add(con);
                        if(!noRelay)
                            ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                    }
                }
                else if(oLas!=null)
                {
if(DEBUG>1) Log.WriteLine("InterGridComms:   Expanding laser \""+((IMyTerminalBlock)oLas).CustomName+"\", "+other.ent.EntityId.ToString());
                    if(other.connectedFrom.Count>0)
                    {
                        radioEnt_t con;
                        {
                            var en=other.connectedFrom.GetEnumerator();
                            en.MoveNext();
                            con=en.Current;
                        }

                        var cLas=con.ent as IMyLaserAntenna;
                        if(cLas==oLas.GetOther())
                        {
if(DEBUG>1) Log.WriteLine("InterGridComms:     ConTo \""+((IMyTerminalBlock)cLas).CustomName+"\", "+con.ent.EntityId.ToString());

                            if(!re.scanMap.Contains(con))
                            {
                                re.scanMap.Add(con);
                                ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                            }
                        }
                    }
                }
                else
                    //Other types can't pass signals
                    continue;

                //Add blocks on same grid
                var other_cg=((IMyCubeBlock)other.ent).CubeGrid;
                if(!ucm_cg.Contains(other_cg)) //No need to check same grid twice
                {
                    foreach(var con in grid2blks[other_cg])
                    {
                        if(!(con.ent is Ingame.IMyRadioAntenna)&&!(con.ent is IMyLaserAntenna))
                            continue;
if(DEBUG>1) Log.WriteLine("InterGridComms:     SameGrid \""+((IMyTerminalBlock)con.ent).CustomName+"\", "+con.ent.EntityId.ToString());

                        if(re.scanMap.Contains(con)) continue;

                        if(((IMyCubeBlock)con.ent).GetUserRelationToOwner(((IMyCubeBlock)other.ent).OwnerId) == MyRelationsBetweenPlayerAndBlock.Enemies)
                            continue;

                        re.scanMap.Add(con);
                        ucm_openset.Enqueue(con);
if(DEBUG>1) Log.WriteLine("InterGridComms:       Ok");
                    }
                    ucm_cg.Add(other_cg);
                }
else if(DEBUG>1) Log.WriteLine("InterGridComms:     SameGrid - already checked");
            }
            re.last_scan_update = clock;
        }

        //////////////////////////////////////////////////////////////////////////////

        public void SendMsg(IMyTerminalBlock sender, string msg)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SendMsg");
            if(msg==null) throw new System.ArgumentException("InterGridComms: null message not supported");
            if(msg.Length>MSG_QUEUE_LIMIT) throw new System.ArgumentException("InterGridComms: message size exceeds queue limit");
            if(!sender.IsWorking || !sender.IsFunctional) return;

            channel_t chan;
            {
                int i=msg.IndexOf('\0');
                if(i==-1 || i==0) throw new System.ArgumentException("InterGridComms: message has empty or missing channel name prefix");
                if(!chanInfo.TryGetValue(msg.Substring(0,i),out chan)) return; //Nobody listening
            }

            radioEnt_t re=radioEnts[sender];
            if(!re.wasOperatingAtLastUpdate) return; //Don't allow power-cycle cheating
            UpdateConnMap(re);
            foreach(var ant in chan.listeners)
                if(ant.ant.IsWorking && ant.ant.IsFunctional && re.connMap.Contains(ant.ant) && radioEnts.ContainsKey(ant.ant))
                {
                    ant.msgs.Enqueue(msg);
                    ant.msgsSize+=msg.Length;

                    //Drop old messages if queue has exceeded size limit
                    while(ant.msgsSize>MSG_QUEUE_LIMIT)
                    {
                        string msg2=ant.msgs.Dequeue();
                        ant.msgsSize-=msg2.Length;
                    }

                    //Trigger PB callback
                    if(ant.msgs.Count>0 && ant.pbcb!=null && !needsPBTriggered.Contains(ant))
                        needsPBTriggered.Add(ant);
                }
        }

        public string RetrieveMsg(IMyTerminalBlock recvr)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - RetrieveMsg");
            antInfo_t ri;
            if(!antInfo.TryGetValue(recvr,out ri) ||
               ri.msgs.Count==0                     )
               return null;

            string msg=ri.msgs.Dequeue();
            ri.msgsSize-=msg.Length;
            check_remove(ri);
            return msg;
        }

        public string GetChannels(IMyTerminalBlock ant)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetChannels");
            antInfo_t ri;
            if(antInfo.TryGetValue(ant,out ri))
               return ri.channels_src;
            return String.Empty;
        }

        public void SetChannels(IMyTerminalBlock ant, string channels)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SetChannels");
            if(channels==null) channels=String.Empty;

            antInfo_t ri=null;
            if(antInfo.TryGetValue(ant,out ri))
            {
                if(channels==ri.channels_src)
                    return; //No change
            }

            string[] channels2=channels.Split('\0');
            HashSet<string> channels3=new HashSet<string>(channels2);
            if(channels3.Contains(String.Empty))
                throw new System.ArgumentException("InterGridComms: listening on channel with empty name not supported");
            foreach(string ch in channels3)
                if(ch.Length > CHAN_NAME_LIMIT)
                    throw new System.ArgumentException("InterGridComms: channel names cannot exceed "+CHAN_NAME_LIMIT.ToString()+" character");

            if(ri==null)
                ri=GetOrMakeAntInfo(ant);

            //Clean up old settings
            foreach(string ch in ri.channels)
            {
                channel_t ci;
                if(!chanInfo.TryGetValue(ch, out ci))
                    continue;
                ci.listeners.Remove(ri);
                if(ci.listeners.Count==0)
                    chanInfo.Remove(ch);
            }

            ri.channels=channels2;
            ri.channels_src=channels;

            foreach(string ch in ri.channels)
            {
                channel_t ci;
                if(!chanInfo.TryGetValue(ch, out ci))
                {
                    ci=new channel_t();
                    chanInfo.Add(ch,ci);
                }
                ci.listeners.Add(ri);
            }

            //Clean up messages that couldn't be received now
            for(int i=ri.msgs.Count;i>0;--i)
            {
                string msg=ri.msgs.Dequeue();
                ri.msgsSize=0;
                if(channels3.Contains(msg.Substring(0,msg.IndexOf('\0'))))
                {   //Still ok
                    ri.msgs.Enqueue(msg);
                    ri.msgsSize+=msg.Length;
                }
            }

            check_remove(ri);
        }

        public Ingame.IMyProgrammableBlock GetPBCallback(IMyTerminalBlock ant)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetPBCallback");
            antInfo_t ri;
            if(antInfo.TryGetValue(ant,out ri))
               return ri.pbcb;
            return null;
        }

        public void SetPBCallback(IMyTerminalBlock ant, Ingame.IMyProgrammableBlock pb)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SetPBCallback");
            //Because we're introducing a method for storing block references, we have to be
            //a little careful here. By only allowing callback to a PB on the same grid
            //as the antenna, we should significantly reduce any exploiting.
            if(ant.CubeGrid != pb.CubeGrid)
                throw new System.ArgumentException("InterGridComms: Callbacks are only supported for Programmable blocks on the same cubegrid as the Antenna");

            antInfo_t ri=GetOrMakeAntInfo(ant);
            ri.pbcb=pb;
        }

        public string GetPBArgument(IMyTerminalBlock ant)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetPBArgument");
            antInfo_t ri;
            if(antInfo.TryGetValue(ant,out ri))
               return ri.pbarg;
            return String.Empty;
        }

        public void SetPBArgument(IMyTerminalBlock ant, string arg)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SetPBArgument");
            antInfo_t ri=GetOrMakeAntInfo(ant);
            ri.pbarg=arg;
        }

        public string GetUnique(IMyTerminalBlock ant)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetUnique");
            antInfo_t ri;
            if(antInfo.TryGetValue(ant,out ri))
               return ri.unique;
            return encode_entityid(ant);
        }

        public string RetrieveScan(IMyTerminalBlock recvr)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - RetrieveScan");
            if(!recvr.IsWorking || !recvr.IsFunctional) return null;
if(DEBUG>1) Log.WriteLine("InterGridComms: RetrieveScan from \""+recvr.CustomName+"\" ("+recvr.EntityId.ToString()+"\"");

            radioEnt_t re=radioEnts[recvr];
            if(!re.wasOperatingAtLastUpdate) return null; //Don't allow power-cycle cheating

            UpdateScanMap(re);
            var sb=new StringBuilder();
            foreach(var other in re.scanMap)
            {
                var oFB=other.ent as IMyFunctionalBlock;
                if(other.ent is IMyLaserAntenna) continue; //Don't include in scan results
                if(!radioEnts.ContainsKey(other.ent)) continue; //Has been deleted since map

                string name="";
                MyRelationsBetweenPlayerAndBlock relat1=MyRelationsBetweenPlayerAndBlock.NoOwnership;
                MyRelationsBetweenPlayerAndBlock relat2=MyRelationsBetweenPlayerAndBlock.NoOwnership;
                if(oFB!=null)
                {
                    if(!oFB.IsWorking || !oFB.IsFunctional) continue; //Check that the object is broadcasting
                    if(oFB is Ingame.IMyRadioAntenna && !((Ingame.IMyRadioAntenna)oFB).IsBroadcasting) continue;

                    name=oFB.CustomName;
                    relat1=recvr.GetUserRelationToOwner(oFB.OwnerId);
                    relat2=oFB.GetUserRelationToOwner(recvr.OwnerId);
                }
                else if(other.ent is IMyCharacter)
                {
                    var oPly=MyAPIGateway.Players.GetPlayerControllingEntity(other.ent);
                    //Check that the player has broadcasting on
                    if(oPly==null || !other.wasOperatingAtLastUpdate) continue;
                    name=other.ent.DisplayName;

                    relat1=recvr.GetUserRelationToOwner(oPly.IdentityId);
                    relat2=oPly.GetRelationTo(recvr.OwnerId);
                    //MyPlayer.GetRelationBetweenPlayers(recvr.OwnerId,oPly.IdentityId);
                }
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - relat -> \""+name+"\" = "+relat1.ToString());
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - relat <- \""+name+"\" = "+relat2.ToString());
                var relat=MyRelationsBetweenPlayerAndBlock.Neutral;
                if(relat1==MyRelationsBetweenPlayerAndBlock.Enemies ||
                   relat2==MyRelationsBetweenPlayerAndBlock.Enemies   )
                   relat=MyRelationsBetweenPlayerAndBlock.Enemies;
                else if( (relat1==MyRelationsBetweenPlayerAndBlock.Owner ||
                          relat1==MyRelationsBetweenPlayerAndBlock.FactionShare) &&
                         (relat2==MyRelationsBetweenPlayerAndBlock.Owner ||
                          relat2==MyRelationsBetweenPlayerAndBlock.FactionShare)   )
                    relat=MyRelationsBetweenPlayerAndBlock.Owner;
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - relat <-> \""+name+"\" = "+relat.ToString());

                if(name.IndexOf('\0')!=-1)
                    name=name.Replace("\0","");
                sb.Append(name);
                sb.Append('\0');

                //Flags
                if(other.ent is Ingame.IMyRadioAntenna) sb.Append('A');
                if(other.ent is Ingame.IMyBeacon) sb.Append('B');
                if(relat==MyRelationsBetweenPlayerAndBlock.Enemies) sb.Append('E');
                if(relat==MyRelationsBetweenPlayerAndBlock.Owner)   sb.Append('F');
                if(relat==MyRelationsBetweenPlayerAndBlock.Neutral) sb.Append('N');
                if(other.ent is IMyCharacter) sb.Append('P');
                sb.Append('\0');

                //Broadcast message
                string msg;
                if(broadcastMessages.TryGetValue(other.ent,out msg))
                    sb.Append(msg);
                sb.Append('\0');

                //Position
                var pos=other.ent.GetPosition();
                sb.AppendFormat("{0:R}\0{1:R}\0{2:R}\0", pos.X, pos.Y, pos.Z);
            }
            sb.Remove(sb.Length-1,1);
            return sb.ToString();
        }

        public void SetBroadcastMessage(IMyTerminalBlock blk, string msg)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - SetBroadcastMessage");
            if(String.IsNullOrEmpty(msg))
            {
                broadcastMessages.Remove(blk);
                return;
            }

            if(msg.Length > CHAN_NAME_LIMIT)
                throw new System.ArgumentException("InterGridComms: broadcast message cannot exceed "+CHAN_NAME_LIMIT.ToString()+" character");
            if(msg.IndexOf('\0')!=-1)
                throw new System.ArgumentException("InterGridComms: Embedded nuls not supported in broadcast message");

            broadcastMessages.Add(blk,msg);
        }

        public string GetBroadcastMessage(IMyTerminalBlock blk)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetBroadcastMessage");
            string msg;
            if(!broadcastMessages.TryGetValue(blk,out msg))
                return null;
            return msg;
        }

        static IGC()
        {
            if(((IMyUtilities)MyAPIUtilities.Static).FileExistsInGlobalStorage("InterGridComms.debug"))
                DEBUG=999;
            else
                DEBUG=0;
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - static IGC");
        }
        
        public IGC()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - IGC");
if(DEBUG>1) Log.WriteLine("InterGridComms: IGC constructor, this="+this.GetHashCode().ToString());
            Instance=this;
            clock=0;
        }
        
        public void setupBlockProperties()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - setupBlockProperties");
            //MyRadioAntenna explicity removes support for MyTerminalBlock, so we have to use
            // MyFunctionalBlock instead
            //Note, we intentionally don't check Instance!=null below because we want to
            // to get the exception if the mod is unloaded (which shouldn't be possible).
            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_SENDMSG,
                null,
                (b,v) => { IGC.Instance.SendMsg(b,v); }
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_RECVMSG,
                (b) => { return IGC.Instance.RetrieveMsg(b); },
                null
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_OPENCHANNELS,
                (b) => { return Instance.GetChannels(b); },
                (b,v) => { IGC.Instance.SetChannels(b,v); }
            );

            AddProperty2<Ingame.IMyProgrammableBlock,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_PBCALLBACK,
                (b) => { return Instance.GetPBCallback(b); },
                (b,v) => { IGC.Instance.SetPBCallback(b,v); }
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_PBARG,
                (b) => { return IGC.Instance.GetPBArgument(b); },
                (b,v) => { IGC.Instance.SetPBArgument(b,v); }
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_UNIQUE,
                (b) => { return IGC.Instance.GetUnique(b); },
                null
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyLaserAntenna>(
                CTRL_SCAN,
                (b) => { return IGC.Instance.RetrieveScan(b); },
                null
            );

            AddProperty2<string,Sandbox.ModAPI.Ingame.IMyRadioAntenna,Sandbox.ModAPI.Ingame.IMyBeacon>(
                CTRL_BCASTMSG,
                (b) => { return IGC.Instance.GetBroadcastMessage(b); },
                (b,v) => { IGC.Instance.SetBroadcastMessage(b,v); }
            );

            AddProperty<bool,Sandbox.ModAPI.Ingame.IMyProgrammableBlock>(
                CTRL_PREFIX,
                (b) => { return IGC.Instance!=null; },
                null
            );
        }

        protected override void UnloadData()
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - UnloadData");
if(DEBUG>1) Log.WriteLine("InterGridComms: UnloadData, this="+this.GetHashCode().ToString());
            Instance=null; //Mark mod unloaded

            antInfo.Clear();
            chanInfo.Clear();
            needsPBTriggered.Clear();
            npbtOld.Clear();
            
            broadcastMessages.Clear();
            radioEnts.Clear();
            grid2blks.Clear();
            updateSeq.Clear();
        }

        //////////////////////////////////////////////////////////////////////////////

        private antInfo_t GetOrMakeAntInfo(IMyTerminalBlock ant)
        {
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - GetOrMakeAntInfo");
            antInfo_t ri;
            if(!antInfo.TryGetValue(ant,out ri))
            {
                ri=new antInfo_t(ant);
                antInfo.Add(ant,ri);
            }
            return ri;
        }

        private void check_remove(antInfo_t ai)
        {   //Check if this antenna can't do anything
if(DEBUG>2) Log.WriteLine("InterGridComms: Enter - check_remove");
            if(ai.channels_src.Length > 0) return;
            if(ai.msgs.Count > 0) return;

            antInfo.Remove(ai.ant);
        }
        
        private static void AddProperty<TValue,TBlock>(string id, Func<IMyTerminalBlock, TValue> Getter, Action<IMyTerminalBlock, TValue> Setter)
        {
            var p=MyAPIGateway.TerminalControls.CreateProperty<TValue,TBlock>(id);
            p.Getter=Getter ?? MakeWriteOnlyPropertyLambda<TValue>(id);
            p.Setter=Setter ?? MakeReadOnlyPropertyLambda<TValue>(id);
            MyAPIGateway.TerminalControls.AddControl<TBlock>(p);
        }

        private static void AddProperty2<TValue,TBlock1,TBlock2>(string id, Func<IMyTerminalBlock, TValue> Getter, Action<IMyTerminalBlock, TValue> Setter)
        {
            AddProperty<TValue,TBlock1>(id,Getter,Setter);
            AddProperty<TValue,TBlock2>(id,Getter,Setter);
        }

        #pragma warning disable //Disable unreachable code warning (return statement is necessary for type computation, but can't be reached due to exception)
        private static Func<IMyTerminalBlock, TValue> MakeWriteOnlyPropertyLambda<TValue>(string id)
        {   return (b) => { throw new System.InvalidOperationException("InterGridComms: \""+id+" does not support reading"); return default(TValue); }; }

        private static Action<IMyTerminalBlock, TValue> MakeReadOnlyPropertyLambda<TValue>(string id)
        {   return (b,v) => { throw new System.InvalidOperationException("InterGridComms: \""+id+" does not support writing"); return; }; }
        #pragma warning restore
        
        public static string encode_entityid(IMyEntity ent)
        {
            string encode="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
            var sb=new StringBuilder();
            for(long x=ent.EntityId;x>0;x/=encode.Length)
                sb.Append(encode[(int)(x%encode.Length)]);
            return sb.ToString();
        }
    }

    public static class Extensions
    {
        //Return end of laser antenna question
        //(Forced to use this work-around because Keen blocked access to Sandbox.Game.Entities.Cube.MyLaserAntenna
        // so can't use GetOther() )
        public static IMyLaserAntenna GetOther(this IMyLaserAntenna ant1)
        {
            var bsd=new BoundingSphereD(ant1.TargetCoords,2.5);
            foreach(var ent2 in MyAPIGateway.Entities.GetEntitiesInSphere(ref bsd))
            {   //Find any cubegrids at the target location
                var cg=ent2 as IMyCubeGrid;
                if(cg==null) continue;
                
                foreach(var sb in cg.GetBlocksInsideSphere(ref bsd))
                {   //Then find any lasant blocks near that location on the grid
                    var ant2=sb.FatBlock as IMyLaserAntenna;
                    if(ant2==null) continue;
                    
                    //Make sure opposing lasant is pointing back at us
                    var gpos=ant1.CubeGrid.WorldToGridInteger(ant2.TargetCoords);
                    if(gpos.IsInsideInclusive(ant1.Min,ant1.Max))
                        return ant2;
                }
            }
            return null;
        }
    }
}
